#!/opt/alt/python37/bin/python3
#coding:utf-8



import getopt
import glob
import os
import re
import shutil
import subprocess
import sys
import syslog
from tempfile import mkstemp

INSTALL_AS_NAME = 'alt-mod-passenger'

PASSENGER_TEMP_DIR = '/var/passenger'
PASSENGER_DIR = '/opt/passenger'
PASSENGER_ARCH = "/usr/share/alt-mod-passenger/archives/passenger.tar.gz"
PASSENGER_LIBS_ARCH = "/usr/share/alt-mod-passenger/archives/libspassenger.tar.gz"

INTERPRETER_PATH = '/opt/alt/ruby21/bin/ruby'
PASSENGER_MOD_PATH = 'buildout/apache2/mod_passenger.so'
PASSENGER_EXEC_PATH = 'bin/passenger-install-apache2-module'

APACHE_CONFIG = "/conf/conf.d/"
CPANEL_APACHE_CONF = "/conf/includes/pre_main_global.conf"
CPANEL_EXTRA_APACHE_CONF = APACHE_CONFIG
DA_APACHE_CONF = "/conf/extra/httpd-includes.conf"
OTHER_CONF = "/conf.d/"
PASSENGER_INCLUDE_CONST = "\n#For passenger\nInclude "
PASSENGER_INCLUDE_PATH = "/etc/"
PASSENGER_INCLUDE_DIRECTIVE = PASSENGER_INCLUDE_CONST + PASSENGER_INCLUDE_PATH
INSTALL_LOG_PATH = "/var/log/alt-mod-passenger_install.log"

NEED_BUILD = False

CPNAME = subprocess.check_output(['cldetect', '--detect-cp-name']).decode().strip()
CAGEFSCTL = '/usr/sbin/cagefsctl'
CAGEFSKELETON = '/usr/share/cagefs-skeleton/bin'
_syslog = lambda x,y: None

class OopsError (RuntimeError): pass


def open_syslog():

    syslog.openlog(INSTALL_AS_NAME, syslog.LOG_PID, syslog.LOG_LOCAL7)
    syslog.syslog(syslog.LOG_INFO, str(sys.argv))
    syslog.syslog(syslog.LOG_INFO, "Control panel detected as '%s'." % CPNAME)

    return syslog.syslog

def clear_folder(path):
    """
    analog of `rm -rf <path>`
    """

    def err_report(f,p,e): sys.stderr.write('%s\n' % e[1])

    if os.path.exists(path):
        for item in glob.glob('%s/*' % path):
            if os.path.isdir(item):
                shutil.rmtree(item, ignore_errors=False, onerror=err_report)
            else:
                try: os.unlink(item)
                except EnvironmentError as e:
                    sys.stderr.write('%s\n' % e)
##
## Commented cz tests are failed on (os.mkdir(path))
##   OSError: [Errno 17] File exists: '/tmp/tmpqt0hq4'
##
#    if os.path.exists(path):
#        shutil.rmtree(path, ignore_errors=False, onerror=lambda f,p,e: sys.stderr.write('%s\n' % e[1]))
#        os.mkdir(path)


def inflate(path, arch_path):
    """
    Extract passenger archive (sources or pre-builded)
    """
    clear_folder(path)
    cmd = ['/bin/tar', '-zxf', arch_path, '-C', path, '--strip-components', '1']

    returncode = subprocess.call(cmd)
    if returncode:
        raise OopsError("Cannot inflate [%d] : %s" % (returncode, ' '.join(cmd)))


def err(msg):
    sys.stderr.write("\033[41mERROR: {}\033[00m\n" .format(msg))
    syslog.syslog(syslog.LOG_ERR, "ERROR: %s" % msg)


def warn(msg):
    sys.stderr.write("\033[91mWARNING: {}\033[00m\n" .format(msg))
    syslog.syslog(syslog.LOG_WARNING, "WARNING: %s" % msg)


def info(msg):
    sys.stderr.write("\033[94mINFO: {}\033[00m\n" .format(msg))
    syslog.syslog(syslog.LOG_INFO, "INFO: %s" % msg)


def install_da():

    build_dir = '/usr/share/alt-mod-passenger/build'

    build_script = '/usr/share/alt-mod-passenger/utils/passenger-build'
    if not os.path.exists(build_script):
        raise OopsError("File %s does not exist" % build_script)

    install_script = "/usr/share/alt-mod-passenger/utils/da-install-passenger"
    if not os.path.exists(install_script):
        raise OopsError("File %s does not exist" % install_script)

    try:
        os.mkdir(build_dir)
    except OSError:
        pass

    inflate(build_dir, PASSENGER_ARCH)
    os.chdir(build_dir)

    install_log = open(INSTALL_LOG_PATH, 'a')
    cmd = [build_script]
    try:
        info("Build passenger...")
        subprocess.check_call(
            cmd,
            stdout=install_log,
            stderr=install_log)
        info("Build passenger... OK")
    except subprocess.CalledProcessError:
        raise SystemExit("Could not compile module with %s, for details see %s" % (build_script, INSTALL_LOG_PATH))

    subprocess.call(install_script)
    clear_folder(build_dir)


def install_without_panel():
    install_script = "/usr/share/alt-mod-passenger/utils/wp-install-passenger"

    if not os.path.exists(install_script):
        raise OopsError("File %s does not exist" % install_script)

    subprocess.call(install_script)

def install():
    """
    Build passenger from sources and install apache directives for each panel
    """

    if CPNAME == "DirectAdmin":
        install_da()
    else:
        install_without_panel()


def delete_da():
    delete_script = "/usr/share/alt-mod-passenger/utils/da-delete-passenger"

    if not os.path.exists(delete_script):
        raise OopsError("File %s does not exist" % delete_script)

    subprocess.call(delete_script)

def delete_without_panel():
    delete_script = "/usr/share/alt-mod-passenger/utils/wp-delete-passenger"

    if not os.path.exists(delete_script):
        raise OopsError("File %s does not exist" % delete_script)

    subprocess.call(delete_script)

def uninstall():
    """
    Uninstall passenger configurations from system files and remove
    builded/extracted targets
    """
    if CPNAME == "DirectAdmin":
        delete_da()
    else:
        delete_without_panel()


def is_cagefs_ok():
    """
    Check if cagefs is fully-functional
    :return: True or False
    """
    return os.path.isdir(CAGEFSKELETON) and os.path.isfile(CAGEFSCTL)


def setup_passenger_cagefs():
    """
    Setup passenger module into cagefs
    """
    if not is_cagefs_ok():
        return
    try:
        cagefs_setup_mount_dir_passenger(PASSENGER_TEMP_DIR)
    except EnvironmentError as e:
        raise OopsError("Could not setup Passenger [%s]" % str(e))


def cagefs_setup_mount_dir_passenger(dir_path, mode=0o755):
    """
    Add CageFS mount point for passenger directories
    :param dir_path: directory path
    :param mode: permissions
    """
    # ensure permissions
    os.chmod(dir_path, mode)
    # add new line if needed
    retcode = subprocess.call([CAGEFSCTL, "--check-mp"],
                              stdout=subprocess.DEVNULL,
                              stderr=subprocess.DEVNULL)
    if not retcode:
        cagefs_mp = open('/etc/cagefs/cagefs.mp', "r+")
        try:
            # check if dir is already present
            if not [x for x in cagefs_mp if x.startswith(dir_path)]:
                # add dir
                cagefs_mp.seek(0, 2)  # 2: seek to the end of file
                print("# next line is added by alt-mod-passenger", file=cagefs_mp)
                print(dir_path, file=cagefs_mp, flush=True)
                # remount
                subprocess.call([CAGEFSCTL, "--remount-all"])
        finally:
            cagefs_mp.close()

def usage():
    print('Usage: %s OPTIONS' % (sys.argv[0],))
    print('')
    print('-i| --install : installs the module')
    print('-d| --delete  : removes the module')
    print('-h| --help    : shows this help')
    print('-p| --path    : path to look for sources. Default path is %s' % (PASSENGER_DIR,))
    print('-c| --config  : path to apache conf file. Default is %s' % (CPANEL_EXTRA_APACHE_CONF,))
    print('--setup-passenger-cagefs : configure passenger with CageFS')


def main(argv):
    """
    Main run function
    """
    opts, args = getopt.getopt(argv, 
        'idhp:c:',
        ['install', 'delete', 'path=', 'config=', 'help', "setup-passenger-cagefs"])

    global _syslog
    _syslog = open_syslog()

    for o, a in opts:
        if o in ['-h', '--help']:
            usage()
            sys.exit(0)
        elif o in ['-i', '--install']:
            install()
        elif o in ['-d', '--delete']:
            uninstall()
        elif o in ['-p', '--path']:
            PASSENGER_DIR = a
        elif o in ['-c', '--config']:
            APACHE_CONFIG = a
        elif o == '--setup-passenger-cagefs':
            setup_passenger_cagefs()


if __name__ == '__main__':
    try:
        main(sys.argv[1:])
    except OopsError as e:
        _syslog(syslog.LOG_ERR, str(e))
        sys.exit(1)
    except Exception as e:
        _syslog(syslog.LOG_ERR, str(e))
        # dump backtrace to stderr for unexpected exception type
        raise
    else:
        _syslog(syslog.LOG_INFO, "No errors.")
        sys.exit(0)
